/*
 * The MIT License (MIT)
 * Copyright (c) 2018 Novel Bits
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 *
 */

#ifndef RELAY_CLIENT_H__
#define RELAY_CLIENT_H__

#include "ble.h"
#include "ble_db_discovery.h"
#include "nrf_sdh_ble.h"
#include <stdint.h>

/**@brief   Macro for defining an relay_client instance.
 *
 * @param   _name   Name of the instance.
 * @hideinitializer
 */

#define BLE_RELAY_CLIENT_BLE_OBSERVER_PRIO 2

#define BLE_RELAY_CLIENT_DEF(_name)       \
  static ble_relay_c_t _name;        \
  NRF_SDH_BLE_OBSERVER(_name##_obs,       \
      BLE_RELAY_CLIENT_BLE_OBSERVER_PRIO, \
      ble_relay_c_on_ble_evt, &_name)


#define OPCODE_LENGTH 1
#define HANDLE_LENGTH 2

/**@brief   Maximum length of data (in bytes) that can be transmitted to the peer by the Nordic UART service module. */
#if defined(NRF_SDH_BLE_GATT_MAX_MTU_SIZE) && (NRF_SDH_BLE_GATT_MAX_MTU_SIZE != 0)
    #define BLE_RELAY_MAX_DATA_LEN (NRF_SDH_BLE_GATT_MAX_MTU_SIZE - OPCODE_LENGTH - HANDLE_LENGTH)
#else
    #define BLE_NUS_MAX_DATA_LEN (BLE_GATT_MTU_SIZE_DEFAULT - OPCODE_LENGTH - HANDLE_LENGTH)
    #warning NRF_SDH_BLE_GATT_MAX_MTU_SIZE is not defined.
#endif


// RELAY service:            E3C30001-411A-4753-A277-4A803A2F67BC
// RELAY action characteristic:   E3C30002-411A-4753-A277-4A803A2F67BC
// RELAY state characteristic:   E3C30003-411A-4753-A277-4A803A2F67BC

// The bytes are stored in little-endian format, meaning the
// Least Significant Byte is stored first
// (reversed from the order they're displayed as)

// Base UUID: E3C30000-411A-4753-A277-4A803A2F67BC
#define BLE_UUID_RELAY_SERVICE_BASE_UUID {{ 0xBC, 0x67, 0x2F, 0x3A, 0x80, 0x4A, 0x77, 0xA2, 0x53, 0x7A, 0x1A, 0x41, 0x00, 0x00, 0xC3, 0xE3 }}

// Service & characteristics UUIDs
#define BLE_UUID_RELAY_SERVICE_UUID 0x0001 //service
#define BLE_UUID_RELAY_ACTION_UUID 0x0002  //characteristic
#define BLE_UUID_RELAY_STATE_UUID 0x0003   //characteristic

/**@brief RELAY_service Client event type. */
typedef enum {
  BLE_RELAY_C_EVT_DISCOVERY_COMPLETE = 1, /**< Event indicating that the RELAY Button Service has been discovered at the peer. */
  BLE_RELAY_C_EVT_STATE_EVT,              /**< Event indicating that the central has received something from a peer. */
  BLE_RELAY_C_EVT_DISCONNECTED            /**< Event indicating that the NUS server has disconnected. */
} ble_relay_client_evt_type_t;

/**@brief Structure containing the handles related to the RELAY Button Service found on the peer. */
typedef struct
{
  uint16_t relay_action_handle; /**< Handle of the RELAY characteristic as provided by the SoftDevice. */
  uint16_t relay_state_handle;  /**< Handle of the RELAY characteristic as provided by the SoftDevice. */
  uint16_t relay_cccd_handle;   /**< Handle of the CCCD of the Relay State characteristic. */
} relay_service_db_t;

/**@brief RELAY Event structure. */
typedef struct
{
  ble_relay_client_evt_type_t evt_type; /**< Type of the event. */
  uint16_t conn_handle;                 /**< Connection handle on which the event occured.*/
  uint16_t max_data_len;
  uint8_t *p_data;
  uint8_t data_len;
  relay_service_db_t peer_db; /**< RELAY Service related handles found on the peer device. This will be filrelay if the evt_type is @ref BLE_RELAY_C_EVT_DISCOVERY_COMPLETE.*/
} ble_relay_client_evt_t;

// Forward declaration of the ble_relay_c_t type.
typedef struct ble_relay_client_s ble_relay_c_t;

/**@brief   Event handler type.
 *
 * @details This is the type of the event handler that should be provided by the application
 *          of this module in order to receive events.
 */
typedef void (*ble_relay_client_evt_handler_t)(ble_relay_c_t *p_relay_client, ble_relay_client_evt_t *p_evt);

/**@brief RELAY Service Client structure. */
struct ble_relay_client_s {
  uint16_t conn_handle;                       /**< Connection handle as provided by the SoftDevice. */
  relay_service_db_t peer_relay_service_db;   /**< Handles related to RELAY Service on the peer*/
  ble_relay_client_evt_handler_t evt_handler; /**< Application event handler to be calrelay when there is an event related to the RELAY service. */
  uint8_t uuid_type;                          /**< UUID type. */
};

/**@brief RELAY Service Client initialization structure. */
typedef struct
{
  ble_relay_client_evt_handler_t evt_handler; /**< Event handler to be calrelay by the RELAY Service Client module whenever there is an event related to the RELAY Service. */
} ble_relay_c_init_t;

/**@brief Function for initializing the RELAY Service client module.
 *
 * @details This function will register with the DB Discovery module. There it registers for the
 *          RELAY Service. Doing so will make the DB Discovery module look for the presence
 *          of a RELAY Service instance at the peer when a discovery is started.
 *
 * @param[in] p_ble_relay_client      Pointer to the RELAY Service client structure.
 * @param[in] p_ble_relay_client_init Pointer to the RELAY Service initialization structure containing the
 *                                          initialization information.
 *
 * @retval    NRF_SUCCESS On successful initialization. Otherwise an error code. This function
 *                        propagates the error code returned by the Database Discovery module API
 *                        @ref ble_db_discovery_evt_register.
 */
uint32_t ble_relay_client_init(ble_relay_c_t *p_ble_relay_client, ble_relay_c_init_t *p_ble_relay_client_init);

/**@brief Function for handling BLE events from the SoftDevice.
 *
 * @details This function will handle the BLE events received from the SoftDevice. If a BLE event
 *          is relevant to the RELAY Service Client module, then it uses it to update interval
 *          variables and, if necessary, send events to the application.
 *
 * @param[in] p_ble_evt     Pointer to the BLE event.
 * @param[in] p_context     Pointer to the RELAY service client structure.
 */
void ble_relay_c_on_ble_evt(ble_evt_t const *p_ble_evt, void *p_context);

/**@brief Function for handling events from the database discovery module.
 *
 * @details Call this function when getting a callback event from the DB discovery module. This
 *          function will handle an event from the database discovery module, and determine if it
 *          relates to the discovery of RELAY Button service at the peer. If so, it will call the
 *          application's event handler indicating that the RELAY Button service has been discovered
 *          at the peer. It also populates the event with the service related information before
 *          providing it to the application.
 *
 * @param[in] p_ble_relay_client Pointer to the RELAY Button client structure.
 * @param[in] p_evt Pointer to the event received from the database discovery module.
 */
void ble_relay_client_on_db_disc_evt(ble_relay_c_t *p_ble_relay_client, const ble_db_discovery_evt_t *p_evt);

/**@brief     Function for assigning a Handles to this instance of relay_client.
 *
 * @details Call this function when a link has been established with a peer to associate this link
 *          to this instance of the module. This makes it  possible to handle several links and
 *          associate each link to a particular instance of this module.
 *
 * @param[in] p_ble_relay_client    Pointer to the RELAY Button client structure instance to associate.
 * @param[in] conn_handle    Connection handle to associate with the given RELAY Button Client Instance.
 * @param[in] p_peer_handles RELAY Button Service handles found on the peer (from @ref BLE_RELAY_C_EVT_DISCOVERY_COMPLETE event).
 *
 */
uint32_t ble_relay_client_handles_assign(ble_relay_c_t *p_ble_relay_client,
    uint16_t conn_handle,
    const relay_service_db_t *p_peer_handles);

/**@brief Function for writing the RELAY2 setting to the connected server.
 *
 * @param[in] p_ble_relay_client Pointer to the RELAY Service client structure.
 * @param[in] setting                  RELAY2 setting to send.
 *
 * @retval NRF_SUCCESS If the staus was sent successfully. Otherwise, an error code is returned.
 */
uint32_t ble_relay_c_action_send(ble_relay_c_t *p_ble_relay_client, uint8_t relay_action);
uint32_t ble_relay_c_state_notif_enable(ble_relay_c_t *p_ble_relay_client);
void relay_send_write(ble_relay_c_t *p_ble_relay_c, uint16_t handle, uint8_t *p_string, uint16_t length);

#endif // RELAY_CLIENT_H__